#[allow(unused_must_use)]
pub mod shell {
    use std::process::Command;
    use std::io::stdin;
    use std::io::stdout;
    use std::io::Write;
    use std::path::Path;
    use std::env;
    use std::str;
    use std::process::Stdio;
    use std::num::Wrapping;
    use users::{get_user_by_uid, get_current_uid};
    use termion::event::Key;
    use termion::input::TermRead;
    use termion::raw::IntoRawMode;
    use termion::{color, cursor, cursor::DetectCursorPos};
    use unicode_segmentation::UnicodeSegmentation;

    mod history;
    use crate::shell::history::*;

    mod autocomplete;
    use crate::shell::autocomplete::*;


    //used for home directory
    extern crate dirs;



    fn process_line(input: &str) -> String {
        let mut commands = input.trim().split(" | ").peekable();
        let mut previous_command = None;
        let mut resultat = String::new();

        while let Some(command) = commands.next()  {
            let mut parts = command.trim().split_whitespace();
            let command = parts.next().unwrap();
            let args = parts;

            match command {
                "cd" => {
                    let new_dir = args.peekable().peek().map_or("/", |x| *x);
                    let root = Path::new(new_dir);
                    if let Err(e) = env::set_current_dir(&root) {
                        eprintln!(" Err: {}", e);
                    }
                    previous_command = None;
                },
                "history" => {
                    //let stdout = Stdio::inherit();
                    let res = get_history();
                    match res {
                        Ok(r) => return r,
                        Err(e) => eprintln!(" Err: {}", e),
                    }
                },
                command => {
                    if commands.peek().is_some() {
                        let stdin  = match previous_command {
                            None => Stdio::inherit(),
                            Some(o) => o,
                        };
                        let stdout = Stdio::piped();
                    
                        let output = Command::new(command)
                            .args(args)
                            .stdout(stdout)
                            .stdin(stdin)
                            .spawn();

                        match output {
                            Ok(o) => previous_command = Some(Stdio::from(o.stdout.unwrap())),
                            Err(e) => {
                                previous_command = None;
                                eprintln!(" Err: {}", e);
                            },
                        };
                    } else {
                        let stdin = match previous_command {
                            None => Stdio::inherit(),
                            Some(o) => o,
                        };
                        let output = Command::new(command)
                            .args(args)
                            .stdin(stdin)
                            .output();

                        let command_result = match output {
                            Ok(o) => o,
                            Err(e) => {
                                eprintln!("\r\nGot error {}", e);
                                return String::from("!");
                            },
                        };

                        let _ = &mut resultat.push_str(str::from_utf8(&command_result.stdout)
                                                    .expect("Could not convert command to str")
                                                    );
                        previous_command = None;
                    }
                },
            };
        }
        return resultat;
    }

    fn build_prompt(curr_number: &i32) -> (String, usize) {
        let mut length: usize = 9;

        length += curr_number.to_string().len();

        let user = String::from(
            get_user_by_uid(get_current_uid())
            .unwrap()
            .name()
            .to_str()
            .unwrap()
            );

        length += user.chars().count();

        let host = String::from(
            gethostname::gethostname()
            .to_str()
            .unwrap()
            );

        length += host.chars().count();

        let current_dir = String::from(
            env::current_dir()
            .unwrap()
            .to_str()
            .unwrap()
            );

        let s = current_dir.split('/');
        let dir_last = String::from(
            s.last()
            .unwrap()
            );

        length += dir_last.chars().count();

        let prompt = String::from(format!("{}{}{}[{}@{} in {}]{} ",
                                          color::Fg(color::LightBlack),
                                          curr_number,
                                          color::Fg(color::LightCyan), 
                                          user, host, dir_last, 
                                          color::Fg(color::Reset)));
        return (prompt, length);
    }

    fn replace_signs(line: &String) -> String {
        let mut ayo = String::from(line);
        if ayo.contains('~') {
            let homedir = dirs::home_dir().unwrap().into_os_string().into_string().unwrap();
            let result = str::replace(&ayo, "~", &homedir);
            ayo = String::from(result);
        }
        
        if ayo.contains("!!") {
            let prev_comm = get_hist_from_number(&get_history_number().unwrap()).unwrap();
            let result = str::replace(&ayo, "!!", &prev_comm);
            ayo = String::from(result);
        }

        return ayo;
    }

    fn handle_input(line: &str) -> String {
        //history callback
        if (line.len() > 0) && (line.starts_with('!'))  {
            let command_found = treat_history_callback(line);
            match command_found {
                Some(c) => {
                    write_to_history(&c);
                    let outp = process_line(&c);
                    return outp;
                    },
                None => return String::from(" "),
            };
        //Regular command
        } else if line.len() > 0 {
            write_to_history(line);
            return process_line(line);
        //Nothing
        } else {
            //Command was empty, new line
            return String::from("");
        }
    }

    fn delete_current_cmd(current_cmd: String) -> String {
        //Deletes currently written command
       
        let mut stdout = stdout().into_raw_mode().unwrap();
        //write!(stdout, "{}", cursor::Save);
        for _c in current_cmd.graphemes(true) {
            write!(stdout, "\x1b[D").unwrap();
            write!(stdout, "\x1b[K").unwrap();
        }
        stdout.flush();
        //write!(stdout, "{}", cursor::Restore);
        return String::new();
    }



    pub fn run_raw() {
        let stdin = stdin();
        let mut stdout = stdout().into_raw_mode().unwrap();
        let mut mycommand = String::new();

        //Used in conjunction with Up arrow to go back in history
        //Resetted by pressing Enter
        let mut current_number = get_curr_history_number();
        let (mut prompt, mut pr_len) = build_prompt(&current_number);

        //Initialize
        write!(stdout, "\r\n SquiShell (sqish)--- \r\n{}", prompt);
        stdout.flush();

        for c in stdin.keys() {

            match c.unwrap() {
                Key::Char('\t') => {
                    let res = autocomplete(&mycommand);
                    mycommand.clear();
                    mycommand.push_str(&res);
                    write!(stdout, "\r\n{}{}", prompt, res);
                    stdout.flush();
                }

                Key::Char('\n') => {
                    current_number = get_curr_history_number();
                    (prompt, pr_len) = build_prompt(&current_number);
                    if (mycommand != String::from("\n")) && (mycommand != String::from("exit")) {
                        let comm = replace_signs(&mycommand);
                        let res = handle_input(&comm);
                        mycommand.clear();
                        //Proper printing
                        for line in res.split('\n') {
                            write!(stdout, "\r\n{}", line);
                        }
                        current_number = get_curr_history_number();
                        (prompt, pr_len) = build_prompt(&current_number);
                        write!(stdout, "{}", prompt).unwrap();
                        stdout.flush();
                    } else if mycommand == String::from("exit") {
                        write!(stdout, "\r\n Sayonara \r\n");
                        break;
                    } else {
                        write!(stdout, "\r\n{}", prompt).unwrap();
                        stdout.flush();
                    }
                },

                Key::Ctrl('q') => {
                    writeln!(stdout, "\r\n Sayonara \r\n");
                    break;
                },

                Key::Char(c) => {
                    mycommand.push(c);
                    write!(stdout, "{}", c).unwrap();
                },

                Key::Backspace => {
                    match mycommand.pop() {
                        Some(_) => {},
                        None => continue,
                    }
                    //Move cursor left
                    write!(stdout, "\x1b[D").unwrap();
                    //Delete one character there
                    write!(stdout, "\x1b[K").unwrap();
                    stdout.flush();
                },

                Key::Up => {
                    mycommand = delete_current_cmd(mycommand);
                    current_number -= 1;
                    let lastcmd = match get_hist_from_number(&current_number) {
                        Some(c) => c,
                        None => continue
                    };
                    mycommand = lastcmd.trim().to_string();
                    //let prompt = build_prompt();
                    write!(stdout, "{}", mycommand);
                },

                Key::Down => {
                    mycommand = delete_current_cmd(mycommand);
                    current_number += 1;
                    let lastcmd = match get_hist_from_number(&current_number) {
                        Some(c) => c,
                        None => continue
                    };
                    mycommand = lastcmd.trim().to_string();
                    //let prompt = build_prompt();
                    write!(stdout, "{}", mycommand);
                },

                Key::Left => {
                    //La taille du prompt est incorrecte à cause des couleurs !
                    let curs_pos = stdout.cursor_pos().unwrap().0;
                    if usize::from(curs_pos) > pr_len {
                        write!(stdout, "{}", cursor::Left(1));
                        stdout.flush();
                    }
                },

                Key::Right => {
                    print!("{}", cursor::Right(1));
                }

                _ => (),
            }

            stdout.flush().unwrap();
        }

    }


    #[cfg(test)]
    mod tests {
        use super::*;

        fn inittests() {
            let filepath = dirs::home_dir().unwrap().join("history.sqish");

            let file = File::create(&filepath).expect("Could not create test history");
            writeln!(&file, "1 ls");
            writeln!(&file, "2 pwd");
        }

        #[test]
        fn test_gethistnumber() {
            inittests();
            let mynumber = get_history_number().unwrap();
            assert_eq!(mynumber, 2);
        }

        #[test]
        fn test_gethistfromnumber() {
            inittests();
            let mycmd = get_hist_from_number(&1).unwrap();
            assert_eq!(mycmd, String::from("ls"));
        }

        #[test]
        fn test_history_callback() {
            inittests();
            let expected = String::from("ls");
            assert_eq!(treat_history_callback("!1"), Some(expected));
        }

        #[test]
        fn test_replace_signs() {
            inittests();
            let homedir = dirs::home_dir().unwrap().into_os_string().into_string().unwrap();
            assert_eq!(replace_signs(String::from("ls ~ && echo !!")), String::from(format!("ls {} && echo pwd", homedir)));
        }

        #[test]
        fn test_writehistline() {
            inittests();
            let towrite = String::from("pwd");
            write_to_history(&towrite);

            //Now check directly against the file
            let filepath = dirs::home_dir().unwrap().join("history.sqish");
            let file = File::open(filepath).expect("Error");
            let mut reader = BufReader::new(file).lines();
            let myline = String::from(reader.last().unwrap().expect("Error"));
            let expected = String::from("3 pwd");

            assert_eq!(myline.trim(), expected.trim()); 
        }
    }
}


